# Installation script for python
import os
import re

from setuptools import find_packages, setup

PACKAGE = "qibolab"


# Returns the version
def get_version():
    """Gets the version from the package's __init__ file
    if there is some problem, let it happily fail"""
    VERSIONFILE = os.path.join("src", PACKAGE, "__init__.py")
    initfile_lines = open(VERSIONFILE).readlines()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(VSRE, line, re.M)
        if mo:
            return mo.group(1)


# load long description from README
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name=PACKAGE,
    version=get_version(),
    description="Quantum hardware module and drivers for Qibo",
    author="The Qibo team",
    author_email="",
    url="https://github.com/qiboteam/qibolab",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"": ["runcards/*.yml", "tests/*.yml"]},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    install_requires=[
        "qibo>=0.1.8",
        "networkx",
        "more-itertools",
    ],
    extras_require={
        "docs": [
            "sphinx",
            "furo",
            "recommonmark",
            "sphinxcontrib-bibtex",
            "sphinx_markdown_tables",
            "nbsphinx",
            "IPython",
            "sphinx_copybutton",
            "furo",
        ],
        "tests": [
            "pytest>=7.2.0",
            "pytest-cov>=4.0.0",
            "pytest-env>=0.8.1",
        ],
        "analysis": [
            "pylint>=2.16.0",
        ],
        # TII system dependencies
        "tiiq": [
            "qblox-instruments==0.7.0",
            "qcodes",
            "pyvisa-py==0.5.3",
            "qm-qua>=1.0.1",
            "qualang-tools>=0.13.1",
        ],
    },
    python_requires=">=3.8.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from dataclasses import dataclass
from functools import cached_property
from typing import Optional

import numpy as np
import numpy.typing as npt

ExecRes = np.dtype([("i", np.float64), ("q", np.float64)])


@dataclass
class AveragedResults:
    """Data structure containing averages of ``ExecutionResults``."""

    i: npt.NDArray[np.float64]
    q: npt.NDArray[np.float64]

    def __add__(self, data):
        i = np.append(self.i, data.i)
        q = np.append(self.q, data.q)

        new_execution_results = self.__class__(i, q)

        return new_execution_results

    def to_dict(self):
        """Serialize output in dict"""

        return {
            "MSR[V]": np.sqrt(self.i**2 + self.q**2),
            "i[V]": self.i,
            "q[V]": self.q,
            "phase[rad]": np.angle(self.i + 1.0j * self.q),
        }

    def __len__(self):
        return len(self.i)


@dataclass
class ExecutionResults:
    """Data structure to deal with the output of :func:`qibolab.platforms.abstract.AbstractPlatform.execute_pulse_sequence`"""

    array: npt.NDArray[ExecRes]
    shots: Optional[npt.NDArray[np.uint32]] = None

    @classmethod
    def from_components(cls, is_, qs_, shots=None):
        ar = np.empty(is_.shape, dtype=ExecRes)
        ar["i"] = is_
        ar["q"] = qs_
        ar = np.rec.array(ar)
        return cls(ar, shots)

    @property
    def i(self):
        return self.array.i

    @property
    def q(self):
        return self.array.q

    def __add__(self, data):
        i = np.append(self.i, data.i, axis=0)
        q = np.append(self.q, data.q, axis=0)

        new_execution_results = self.__class__.from_components(i, q)

        return new_execution_results

    @cached_property
    def measurement(self):
        """Resonator signal voltage mesurement (MSR) in volts."""
        return np.sqrt(self.i**2 + self.q**2)

    @cached_property
    def phase(self):
        """Computes phase value."""
        phase = np.angle(self.i + 1.0j * self.q)
        return phase
        # return signal.detrend(np.unwrap(phase))

    @cached_property
    def ground_state_probability(self):
        """Computes ground state probability"""
        return 1 - np.mean(self.shots)

    def to_dict_probability(self, state=1):
        """Serialize probabilities in dict.
        Args:
            state (int): if 0 stores the probabilities of finding
                        the ground state. If 1 stores the
                        probabilities of finding the excited state.
        """
        if state == 1:
            return {"probability": 1 - self.ground_state_probability}
        elif state == 0:
            return {"probability": self.ground_state_probability}

    def compute_average(self):
        """Perform average over i and q"""
        return AveragedResults(self.array.i.mean(), self.array.q.mean())

    def to_dict(self, average=True):
        """Serialize output in dict.
        Args:
            average (bool): If `True` returns a dictionary of the form
                            {'MSR[V]' : v, 'i[V]' : i, 'q[V]' : q, 'phase[rad]' : phase}.
                            Where each value is averaged over the number shots. If `False`
                            all the values for each shot are saved.
        """
        results = self.compute_average() if average else self

        return {
            "MSR[V]": np.sqrt(results.i**2 + results.q**2),
            "i[V]": results.i,
            "q[V]": results.q,
            "phase[rad]": np.angle(results.i + 1.0j * results.q),
        }

    def __len__(self):
        return len(self.i)



import tempfile
from abc import ABC, abstractmethod
from pathlib import Path

from qibolab.paths import user_folder


class AbstractInstrument(ABC):
    """
    Parent class for all the instruments connected via TCPIP.

    Args:
        name (str): Instrument name.
        address (str): Instrument network address.
    """

    def __init__(self, name, address):
        self.name: str = name
        self.address: str = address
        self.is_connected: bool = False
        self.signature: str = f"{type(self).__name__}@{address}"
        # create local storage folder
        instruments_data_folder = user_folder / "instruments" / "data"
        instruments_data_folder.mkdir(parents=True, exist_ok=True)
        # create temporary directory
        self.tmp_folder = tempfile.TemporaryDirectory(dir=instruments_data_folder)
        self.data_folder = Path(self.tmp_folder.name)

    @abstractmethod
    def connect(self):
        raise NotImplementedError

    @abstractmethod
    def setup(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def start(self):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError

    @abstractmethod
    def disconnect(self):
        raise NotImplementedError

    def play(self, *args, **kwargs):
        """Play a pulse sequence and retrieve feedback.

        Returns:
            (dict) mapping the serial of the readout pulses used to
            the acquired :class:`qibolab.result.ExecutionResults` object.
        """
        raise NotImplementedError(f"Instrument {self.name} does not support play.")

    def sweep(self, *args, **kwargs):
        """Play a pulse sequence while sweeping one or more parameters.

        Returns:
            (dict) mapping the serial of the readout pulses used to
            the acquired :class:`qibolab.result.ExecutionResults` object.
        """
        raise NotImplementedError(f"Instrument {self.name} does not support sweep.")


class LocalOscillator(AbstractInstrument):
    """Abstraction for local oscillator instruments.

    Local oscillators are used to upconvert signals, when
    the controllers cannot send sufficiently high frequencies
    to address the qubits and resonators.
    """

    def play(self, *args, **kwargs):
        """Local oscillators do not play pulses."""

    def sweep(self, *args, **kwargs):
        """Local oscillators do not play pulses."""


class InstrumentException(Exception):
    def __init__(self, instrument: AbstractInstrument, message: str):
        header = f"InstrumentException with {instrument.signature}"
        full_msg = header + ": " + message
        super().__init__(full_msg)
